

/**
 * Static class with methods that return an int value casted
 * from another type. Cast to and from the same type acts as
 * a JRelix valid range checker.
 *
 * @since jRelCaNu 2002-07
 */


import java.util.*;


public class CastToInt implements Constants, TypeConstants {  
    static private final int RESULT_TYPE_DC = INT_DC;
    static private final int RESULT_TYPE_DK = INT_DK;
    public static int assertValidValue(int castFromInteger) 
        throws InterpretError {
        return value(castFromInteger);
    };

    /**
     * Checks to see if one JRelix type is castable to another.
     * We must provide a value() method for each castable type.
     *
     * @return  a boolean true if the RHS type can be casted.
     * @since jRelCaNu
     */
    public static boolean domCompatible(int typeRHS) {
        switch (typeRHS) {
        case RELIX_VOID:
            return false;

        case BOOLEAN:
        case SHORT:
        case INTEGER:
        case LONG:
        case FLOAT:
        case DOUBLE:
        case NUMERIC:
        case STRING:
        case TEXT:
        case DC:	//added by Yi Zheng Sept 28, 2002
        case DK:	//added by Yi Zheng Sept 28, 2002
            return true;

        case STMT:
        case EXPR:
        case IDLIST:
        case COMP:
   //     case DC:	//Yi Zheng Sept 28, 2002
   //     case DK:  //Yi Zheng Sept 28, 2002
        case RELATION:
        case VIEW:
        case COMPUTATION:
        case ATTRIBUTE:
        case UNIVERSAL:
        default:
            return false;
        }
    }

    /**
     * Returns an int value from a string.
     *
     * @return  an integer corresponding to the first numeric
     *          string in the entire castFromStrn.
     *          Uses the standard Java methods to convert the
     *          numeric string to a number.
     * @since jRelCaNu
     */
    public static int value(String castFromStrn) 
        throws InterpretError {
        if (castFromStrn.equals(STRING_DK))
            return RESULT_TYPE_DK;
        else if (castFromStrn.equals(STRING_DC))
            return RESULT_TYPE_DC;
        int retVal = value(CastToLong.value(castFromStrn));
        if (retVal == RESULT_TYPE_DC)
            throw new InterpretError(
                    "CastToInt.value: error: Can't cast to RESULT_TYPE_DC.");
        else if (retVal == RESULT_TYPE_DK)
            throw new InterpretError(
                    "CastToInt.value: error: Can't cast to RESULT_TYPE_DK.");
        else 
            return retVal;
    }

    public static int value(byte castFromBool) 
        throws InterpretError {
        if (castFromBool == (byte) BOOL_DK)
            return RESULT_TYPE_DK;
        else if (castFromBool == (byte) BOOL_DC)
            return RESULT_TYPE_DC;
        else if (castFromBool == (byte) BOOL_FALSE) 
            return 0;
        else return 1;
    }

    public static int value(long castFromLong) 
        throws InterpretError {
        if (castFromLong == LONG_DC)
            return RESULT_TYPE_DC;
        else if (castFromLong == LONG_DK)
            return RESULT_TYPE_DK;
        else if ((int) castFromLong == RESULT_TYPE_DC)
            throw new InterpretError(
                    "CastToInt.value: error: Can't cast to RESULT_TYPE_DC.");
        else if ((int) castFromLong == RESULT_TYPE_DK)
            throw new InterpretError(
                    "CastToInt.value: error: Can't cast to RESULT_TYPE_DK.");
        else 
            return ((int) castFromLong);
    }

    public static int value(int castFromInt) 
        throws InterpretError {
        if (castFromInt == INT_DC)
            return RESULT_TYPE_DC;
        else if (castFromInt == INT_DK)
            return RESULT_TYPE_DK;
        else if ((int) castFromInt == RESULT_TYPE_DC)
            throw new InterpretError(
                    "CastToInt.value: error: Can't cast to RESULT_TYPE_DC.");
        else if ((int) castFromInt == RESULT_TYPE_DK)
            throw new InterpretError(
                    "CastToInt.value: error: Can't cast to RESULT_TYPE_DK.");
        else 
            return ((int) castFromInt);
    }

    public static int value(short castFromShort) 
        throws InterpretError {
        if (castFromShort == SHORT_DC)
            return RESULT_TYPE_DC;
        else if (castFromShort == SHORT_DK)
            return RESULT_TYPE_DK;
        else if ((int) castFromShort == RESULT_TYPE_DC)
            throw new InterpretError(
                    "CastToInt.value: error: Can't cast to RESULT_TYPE_DC.");
        else if ((int) castFromShort == RESULT_TYPE_DK)
            throw new InterpretError(
                    "CastToInt.value: error: Can't cast to RESULT_TYPE_DK.");
        else 
            return ((int) castFromShort);
    }

    public static int value(float castFromFloat) 
        throws InterpretError {
        try {
            if (castFromFloat == FLOAT_DC)
                return RESULT_TYPE_DC;
            else if (castFromFloat == FLOAT_DK)
                return RESULT_TYPE_DK;
            else if ((int) castFromFloat == RESULT_TYPE_DC)
                throw new InterpretError(
                        "CastToInt.value: error: Can't cast to RESULT_TYPE_DC.");
            else if ((int) castFromFloat == RESULT_TYPE_DK)
                throw new InterpretError(
                        "CastToInt.value: error: Can't cast to RESULT_TYPE_DK.");
            else if (Float.isInfinite(castFromFloat))
                throw new InterpretError(
                        "CastToInt.value: error: Can't cast to +/- Infinity.");
            else if (Float.isNaN(castFromFloat))
                throw new InterpretError(
                        "CastToInt.value: error: Can't cast NaN.");
            else {
                return ((int) castFromFloat);
            }
        } catch (Exception e) {
            throw new InterpretError(
                    "CastToInt.value(): error: Can't cast from castFromFloat:" + castFromFloat + " with JAVA cast to:" + (new Integer((int) castFromFloat)).toString() + " with error: " + e);
        }
    }

    public static int value(double castFromDouble)
        throws InterpretError {
        try {
            if (castFromDouble == DOUBLE_DC)
                return RESULT_TYPE_DC;
            else if (castFromDouble == DOUBLE_DK)
                return RESULT_TYPE_DK;
            else if ((int) castFromDouble == RESULT_TYPE_DC)
                throw new InterpretError(
                        "CastToInt.value: error: Can't cast to RESULT_TYPE_DC.");
            else if ((int) castFromDouble == RESULT_TYPE_DK)
                throw new InterpretError(
                        "CastToInt.value: error: Can't cast to RESULT_TYPE_DK.");
            else if (Double.isInfinite(castFromDouble))
                throw new InterpretError(
                        "CastToInt.value: error: Can't cast to +/- Infinity.");
            else if (Double.isNaN(castFromDouble))
                throw new InterpretError(
                        "CastToInt.value: error: Can't cast NaN.");
            else {
                return ((int) castFromDouble);
            }
        } catch (Exception e) {
            throw new InterpretError(
                    "CastToInt.value(): error: Can't cast from castFromDouble:" + castFromDouble + " with JAVA cast to:" + (new Integer((int) castFromDouble)).toString() + " with error: " + e);
        }
    }

    public static int value(number castFromNumber) 
        throws InterpretError {
                return value(CastToLong.value(castFromNumber));
    }
}
