

/**
 * Static class with methods that return a short value casted
 * from another type. Cast to and from the same type acts as
 * a JRelix valid range checker.
 *
 * @since jRelCaNu 2002-07
 */


import java.util.*;


public class CastToShort implements Constants, TypeConstants {  
    static private final short RESULT_TYPE_DC = (short) SHORT_DC;
    static private final short RESULT_TYPE_DK = (short) SHORT_DK;
    public static short assertValidValue(short castFromShort) 
        throws InterpretError {
        return value(castFromShort);
    };

    /**
     * Checks to see if one JRelix type is castable to another.
     * We must provide a value() method for each compatible type.
     *
     * @return  a boolean true if the RHS type can be casted.
     * @since jRelCaNu
     */
    public static boolean domCompatible(int typeRHS) {
        switch (typeRHS) {
        case RELIX_VOID:
            return false;

        case BOOLEAN:
        case SHORT:
        case INTEGER:
        case LONG:
        case FLOAT:
        case DOUBLE:
        case NUMERIC:
        case STRING:
        case TEXT:
        case DK:	//added by Yi Zheng Sept 28, 2002
        case DC:	//added by Yi Zheng Sept 28, 2002
            return true;

        case STMT:
        case EXPR:
        case IDLIST:
        case COMP:
 //       case DC:
 //       case DK:
        case RELATION:
        case VIEW:
        case COMPUTATION:
        case ATTRIBUTE:
        case UNIVERSAL:
        default:
            return false;
        }
    }

    /**
     * Returns a short value from a string.
     *
     * @return  a short corresponding to the first numeric
     *          string in the entire castFromStrn.
     *          Uses the standard Java methods to convert the
     *          numeric string to a number.
     * @since jRelCaNu
     */
    public static short value(String castFromStrn) 
        throws InterpretError {
        if (castFromStrn.equals(STRING_DK))
            return RESULT_TYPE_DK;
        else if (castFromStrn.equals(STRING_DC))
            return RESULT_TYPE_DC;
        short retVal = value(CastToLong.value(castFromStrn));
        if (retVal == RESULT_TYPE_DC)
            throw new InterpretError(
                    "CastToShort.value: error: Can't cast to RESULT_TYPE_DC.");
        else if (retVal == RESULT_TYPE_DC)
            throw new InterpretError(
                    "CastToShort.value: error: Can't cast to RESULT_TYPE_DK.");
        else 
            return retVal;
    }

    public static short value(byte castFromBool) 
        throws InterpretError {
        if (castFromBool == (byte) BOOL_DK)
            return RESULT_TYPE_DK;
        else if (castFromBool == (byte) BOOL_DC)
            return RESULT_TYPE_DC;
        else if (castFromBool == (byte) BOOL_FALSE) 
            return 0;
        else return 1;
    }

    public static short value(long castFromLong) 
        throws InterpretError {
        if (castFromLong == LONG_DC)
            return RESULT_TYPE_DC;
        else if (castFromLong == LONG_DK)
            return RESULT_TYPE_DK;
        else if ((short) castFromLong == RESULT_TYPE_DC)
            throw new InterpretError(
                    "CastToShort.value: error: Can't cast to RESULT_TYPE_DC.");
        else if ((short) castFromLong == RESULT_TYPE_DK)
            throw new InterpretError(
                    "CastToShort.value: error: Can't cast to RESULT_TYPE_DK.");
        else 
            return ((short) castFromLong);
    }

    public static short value(int castFromInt) 
        throws InterpretError {
        if (castFromInt == INT_DC)
            return RESULT_TYPE_DC;
        else if (castFromInt == INT_DK)
            return RESULT_TYPE_DK;
        else if ((short) castFromInt == RESULT_TYPE_DC)
            throw new InterpretError(
                    "CastToShort.value: error: Can't cast to RESULT_TYPE_DC.");
        else if ((short) castFromInt == RESULT_TYPE_DK)
            throw new InterpretError(
                    "CastToShort.value: error: Can't cast to RESULT_TYPE_DK.");
        else 
            return ((short) castFromInt);
    }

    public static short value(short castFromShort) 
        throws InterpretError {
        if (castFromShort == SHORT_DC)
            return RESULT_TYPE_DC;
        else if (castFromShort == SHORT_DK)
            return RESULT_TYPE_DK;
        else if ((short) castFromShort == RESULT_TYPE_DC)
            throw new InterpretError(
                    "CastToShort.value: error: Can't cast to RESULT_TYPE_DC.");
        else if ((short) castFromShort == RESULT_TYPE_DK)
            throw new InterpretError(
                    "CastToShort.value: error: Can't cast to RESULT_TYPE_DK.");
        else 
            return ((short) castFromShort);
    }

    public static short value(float castFromFloat) 
        throws InterpretError {
        try {
            if (castFromFloat == FLOAT_DC)
                return RESULT_TYPE_DC;
            else if (castFromFloat == FLOAT_DK)
                return RESULT_TYPE_DK;
            else if ((short) castFromFloat == RESULT_TYPE_DC)
                throw new InterpretError(
                        "CastToShort.value: error: Can't cast to RESULT_TYPE_DC.");
            else if ((short) castFromFloat == RESULT_TYPE_DK)
                throw new InterpretError(
                        "CastToShort.value: error: Can't cast to RESULT_TYPE_DK.");
            else if (Float.isInfinite(castFromFloat))
                throw new InterpretError(
                        "CastToShort.value: error: Can't cast to +/- Infinity.");
            else if (Float.isNaN(castFromFloat))
                throw new InterpretError(
                        "CastToShort.value: error: Can't cast NaN.");
            else {
                return ((short) castFromFloat);
            }
        } catch (Exception e) {
            throw new InterpretError(
                    "CastToShort.value(): error: Can't cast from castFromFloat:" + castFromFloat + " with JAVA cast to:" + (new Short((short) castFromFloat)).toString() + " with error: " + e);
        }
    }

    public static short value(double castFromDouble)
        throws InterpretError {
        try {
            if (castFromDouble == DOUBLE_DC)
                return RESULT_TYPE_DC;
            else if (castFromDouble == DOUBLE_DK)
                return RESULT_TYPE_DK;
            else if ((short) castFromDouble == RESULT_TYPE_DC)
                throw new InterpretError(
                        "CastToShort.value: error: Can't cast to RESULT_TYPE_DC.");
            else if ((short) castFromDouble == RESULT_TYPE_DK)
                throw new InterpretError(
                        "CastToShort.value: error: Can't cast to RESULT_TYPE_DK.");
            else if (Double.isInfinite(castFromDouble))
                throw new InterpretError(
                        "CastToShort.value: error: Can't cast to +/- Infinity.");
            else if (Double.isNaN(castFromDouble))
                throw new InterpretError(
                        "CastToShort.value: error: Can't cast NaN.");
            else {
                return ((short) castFromDouble);
            }
        } catch (Exception e) {
            throw new InterpretError(
                    "CastToShort.value(): error: Can't cast from castFromDouble:" + castFromDouble + " with JAVA cast to:" + (new Short((short) castFromDouble)).toString() + " with error: " + e);
        }
    }

    public static short value(number castFromNumber) 
        throws InterpretError {
                return value(CastToLong.value(castFromNumber));
    }
}
