 /**
 * Class Computation.
 *
 * The constructor handles a computation declaration.
 * Method applyIjoin() handles the application of the computation 
 *    with a relation.
 * Method applySelect() handles the application of the computation
 *    with a select statement.
 *
 * Notes on the use of Environments.
 * We use the discipline of 'static binding' (see Thesis).
 * 1) 'declEnv' stores a pointer to the environment in which the
 *    computation is declared (code is in constructor).
 * 2) 'declEnv' is used as the parent of the environment ('env') in which 
 *    the computation is applied (code is in method initializeEnvironment()
 *    which is called by applyIjoin() and applySelect()).
 *
 * @author Patrick Baker
 */
import java.util.*;
import java.io.*;

   /***************************************************************/
   /*                        CLASS COMPBLOCK                      */
   /***************************************************************/

/**
 * Class CompBlock is used to store information about an "alt" block
 * in a convenient way. This includes
 * blockType         The block type of each 'alt' block, 
 * blockTree         pointer to jjTree for this block
 * statements        Array of pointers to statments contained in the block
 * numStatements     Number of statements in this block
 * redop		true if this block is decorated with "redop" and thus can be used as associative commutative operator
 * funop		true if this block is decorated with "funop" and thus can be used as ordered operator
 */
class CompBlock implements Constants, TypeConstants
{     /***************************************************************/
   /*                           FIELDS                            */
   /***************************************************************/
//Debugging:
private boolean DEBUG=false;

   public int blockType;           //Type of the block
   public SimpleNode[] statements; //Statements contained in the block
   public int numStatements;       //Number of statements in this block
   public SimpleNode blockTree;    //pointer to jjTree for this block
  
//addition by Yi Zheng July 8, 2002
   public boolean redop = false;  //true if can be used as associative commutative operator
   public boolean funop = false; // true if can be used as ordered operator
//end of addition by Yi Zheng July 8, 2002

   private final int INPUT = 0;
   private final int OUTPUT = 1;

public CompBlock(SimpleNode block, Environment env, SimpleNode tree)
   throws InterpretError, SemanticCheckError
   {  
	//added by Yi Zheng July 8, 2002
	if(block.isBitSet(0))
		redop = true;
	else if(block.isBitSet(1))
		funop = true;

	//end of addition July 8, 2002

	numStatements = block.jjtGetNumChildren();

	  statements = new SimpleNode[numStatements];
	  blockTree = block;

	  //Analyze each statement of the computation
	  // 1) to help deduce the block type of 'alt' block
	  // 2) to do some semantic checking
	  for(int i=0; i<numStatements; i++)
	  {  statements[i] = (SimpleNode) block.jjtGetChild(i);
		SemanticChecker.SC_Computation(statements[i], env);
		 findBlockType(statements[i], env);
	  }
	  
	  //Now calculate the block type from the inoutStatus of the parameters
	  blockType = 0; //?? how does this distinguish between cases where all parameters are outputs? (constructor)
	  					//BUG clue! REVISIT Yi Zheng March 30, 2002
	  SimpleNode parameters = (SimpleNode) tree.jjtGetChild(1);
	  if (parameters.type != OP_PARALIST)
	  {  return;  //There are no parameters to this computation
	  }           //Just set block type to 0;
	  int numParams = parameters.jjtGetNumChildren();
	  for(int i=0; i<numParams; i++)
	  {  ParamInfo pi;
		 String paramName = ((SimpleNode) parameters.jjtGetChild(i)).name;
		 pi = env.lookupParam(paramName, false);
		 //added by Yi Zheng March 30, 2002
		 if(pi == null)
		 	throw new InterpretError("parameter not found in env");
		 if(pi.inoutStatus == INPUT) 
//		if(pi.inoutStatus == OUTPUT)
		 {  blockType |= 1 << i;
		 }
	  }

   } /* Constructor */   
   /**
	* Provided for debugging
	*/
   public void dump(int blocknumber)
   {  Global.println("block "+blocknumber+": block type is "+blockType+
						 ": numStatements is "+numStatements);
   } 
   
   //added by Yi Zheng March 31, 2002  
   public void dump()
   { Global.println("block type is " + blockType + " numStatements is " +
   	numStatements);
   }
   
   /**
	* This method implements the block algorithm on one stmt of the comp.
	* (See thesis for a description of the algorithm.)
	*/
   private void findBlockType(SimpleNode stmt, Environment env)
   throws InterpretError
   {  switch(stmt.type)
	  {  case OP_ASSIGNMENT: //REVISIT: case with 4 children
			updateStatus(INPUT, (SimpleNode) stmt.jjtGetChild(1), env);
			
			if(stmt.opcode == OP_ASSIGN)
			{
				 updateStatus(OUTPUT, (SimpleNode) stmt.jjtGetChild(0), env);
			}
			else //OP_ASSIGNADD
			{ updateStatus(INPUT, (SimpleNode) stmt.jjtGetChild(0), env);
			}
	//		updateStatus(INPUT, (SimpleNode) stmt.jjtGetChild(1), env);//moved up by Yi Zheng June 2, 2002
			break;
		 case OP_COMMAND :
			// Weizhong: come back later;
//			throw new 
//			InterpretError("Commands not implemented in computations");

			//added by Yi Zheng April 8, 2002
 			switch(stmt.opcode)
 			{
      		  case OP_HELP : 
		      case OP_QUIT : 
		      case OP_TRACE : 
		      case OP_INPUT : 
		      case OP_SSD : 
		      case OP_SSR : 
		      case OP_DEBUG : 
		      case OP_BATCH : 
		      case OP_EXPERT : 
		      case OP_TIME : 
		      case OP_DELD : 
		      case OP_SD : 
		      case OP_SRD : 
		      case OP_PRINT : 
		      case OP_EVENTON : 
		      case OP_EVENTOFF : 
		      case OP_UNDO:

		      		break;
		      case OP_DELR : {
		      	updateStatus(INPUT, (SimpleNode) stmt.jjtGetChild(0), env);
				break;
		      }
		      case OP_PR : {
		      	updateStatus(INPUT, (SimpleNode) stmt.jjtGetChild(0), env);
				break;
		      }
		      case OP_SC : {
		      	SimpleNode nd = SimpleNode.jjtCreateNode("Identifier", OP_IDENTIFIER, OP_IDENTIFIER, stmt.name);
		      	updateStatus(INPUT, nd, env);
				break;
		      }
		      case OP_SR : {
			if(stmt.jjtGetNumChildren() > 0) {
		      	SimpleNode nd = SimpleNode.jjtCreateNode("Identifier", OP_IDENTIFIER, OP_IDENTIFIER, stmt.name);
		      	updateStatus(INPUT, nd, env);}
				break;
		      }
		      default : {
				throw new InterpretError("No such command");
		      }
		      }			
			break; 
		 case OP_DECLARATION :
			switch(stmt.opcode) 
			{  case OP_COMPUTATION :
				  //Use new Relation() instead of new Computation() so that
				  //the block algorithm will not run. It should only run
				  //when the computation is called. We save
				  //'stmt' in the 'tree' field so that we can call
				  //new Computation() later (See updateStatus())
				  Relation c = new Relation(); //nothing is initialized here Yi Zheng March 30, 2002

				  SimpleNode tempNode;
				  tempNode = (SimpleNode)stmt.jjtGetChild(0);
				  while (tempNode.name == null) {
				    tempNode = (SimpleNode)
				    tempNode.jjtGetChild(0);
				  };
				  c.name = tempNode.name;

//				  c.name = ((SimpleNode) stmt.jjtGetChild(0)).name;
				  c.rvc = COMPUTATION;
				  c.tree = stmt;
				  env.put(c);
				//  if(DEBUG)
				//  {
				//  	System.out.println("In findBlockType:OP_COMPUTATION: sub comp");
				//  	System.out.println("comp: " + c.name);
				//  }
				  break;
			   case OP_DOMAIN :
			   case OP_LET :
				  //Domain and virtual domain declarations do not 
				  //have input or output domains
				  //REVISIT: really? Yi Zheng May 15, 2002
				  break;
			   case OP_RELATION :
				  //Look at optional initialization part for an identifier
				  if (stmt.jjtGetNumChildren() == 3)
				  { updateStatus(INPUT,(SimpleNode)stmt.jjtGetChild(2),env);
				  }
				  break;
			   case OP_VIEW :
				  updateStatus(INPUT, (SimpleNode) stmt.jjtGetChild(1), env);
				  break;
			}
			break;
		 case OP_STATEMENT :
			switch(stmt.opcode) 
			{  case OP_CONDITIONAL :
				  SimpleNode ifNode, thenNode, elseNode;
				  ifNode = (SimpleNode) stmt.jjtGetChild(0);
				  thenNode = (SimpleNode) stmt.jjtGetChild(1);
				  updateStatus(INPUT, ifNode, env);
				  findBlockType(thenNode, env);
				  if (stmt.jjtGetNumChildren() == 3)
				  {  //There is an else clause
					 elseNode = (SimpleNode) stmt.jjtGetChild(2);
					 findBlockType(elseNode, env);
				  }
				  break;
			   case OP_STMTBLOCK :
				  int numChildren = stmt.jjtGetNumChildren();
				  for (int i=0; i<numChildren; i++)
				  {   findBlockType((SimpleNode) stmt.jjtGetChild(i), env);
				  }
				  break;
			   case OP_COMPCALL : //added by Yi Zheng May 20, 2002
			   {
			   		if(!stmt.isBitSet(0)) //no call parameter
			   			break;
			   		SimpleNode callParamList = (SimpleNode) stmt.jjtGetChild(1);
			   		int numParam = callParamList.jjtGetNumChildren();
			   		for (int i = 0; i < numParam; i++)
			   		{
			   			SimpleNode callParam = (SimpleNode) callParamList.jjtGetChild(i);
						if(!callParam.isBitSet(0))
							updateStatus(INPUT, (SimpleNode)callParam.jjtGetChild(0), env);
			   		}
			   		break;
			   }
			   case OP_SEQUENCE :
			   case OP_PARALLEL :
			   case OP_CHOICE :
			   case OP_FORLOOP :
			   case OP_WHILELOOP :
			   case OP_EXIT :
			   case OP_DEADLOCK :
			   case OP_EXEC :
				  throw new InterpretError
						("Statement not implemented in computations");
			   default : // Parser should prevent flow from reaching here
				  throw new InterpretError("Unknown statement type");
			}
			break;
		 case OP_UPDATE :{
			//REVISIT: Last section to be filled in ...
			//added by Yi Zheng May 15, 2002
			
			//there must be an identifier
			updateStatus(INPUT, (SimpleNode) stmt.jjtGetChild(0), env);
			
			//if there are 4 children must be add or delete
			int numChild = stmt.jjtGetNumChildren();
			if (numChild == 4)
				updateStatus(INPUT, (SimpleNode) stmt.jjtGetChild(3), env);
			else if (numChild == 3) //update..change..using
			{
				updateStatus(INPUT, (SimpleNode) stmt.jjtGetChild(2), env);
				
				//added by Yi Zheng June 14, 2002
				if(stmt.isBitSet(0)) //statementlist is present
				{
					SimpleNode stmtList = (SimpleNode) stmt.jjtGetChild(1);
					for(int i = 0; i < stmtList.jjtGetNumChildren(); i++)
						findBlockType((SimpleNode)stmtList.jjtGetChild(i), env);
				}
				
				//end of addition June 14, 2002
			}
			else if (numChild == 2) //update .. add / delete
				updateStatus(INPUT, (SimpleNode) stmt.jjtGetChild(1), env);
			else throw new InterpretError("wrong number of children for update");
			
			//end of addition May 15, 2002
			break;
		}
		 default :
			throw new
			InterpretError("Unknown statement type");
	  }
   } /* findBlockType() */   
   /**
	* This method is called by findBlockType().
	* Special action is taken when a computation is called. The
	* code of the called computation is checked incase it touches
	* the formal parameters of this computation. This all happens 
	* automatically by calling the constructor for the computation. 
	* Also, Project clauses do not contribute to the block algorithm
	* and are ignored
	*/
   private void updateStatus(int newStat, SimpleNode node, Environment env)
   throws InterpretError
   {  if (node.type == OP_IDENTIFIER)
	  {  String idname = node.name;
		 IDInfo id = env.lookup(idname, true);
		 if (id == null)
		 {  return;  //This identifier does not exist yet.
		 }           //A dynamic bind will be done at run time.
		if(id instanceof ParamInfo) //added by Yi Zheng May 15, 2002
		{ 
		 if (id.inoutStatus == RELIX_VOID)
		 {  id.inoutStatus = newStat;
		 }
		 
		 if(newStat == INPUT) 
		 {  //Need special code to run a computation..
			id = env.lookup(idname, false); 
			if (id!=null && id.kind==REL)
			{ 
//				if(DEBUG)
//					System.out.println("computation nested inside");
					
			 Relation r = ((RelInfo) id).rel;
			   if (r.rvc == COMPUTATION)
			   {  //Just create a new computation
				  //This will cause the block algorithm to run on it
				  Computation c = new Computation(r.tree, env);
			   }
			}
		 }
		}
	  }
	  else
	  {  //Do not consider project nor select clauses.
	     //we still need to consider  OP_SELECT
	     //Yi Zheng May 15, 2002
	     if(node.type != OP_PROJECTOR)
		 { for(int i=0; i<node.jjtGetNumChildren(); i++)
		   { updateStatus(newStat, (SimpleNode) node.jjtGetChild(i), env);
		   }
		 }
//		 else if (node.type == OP_SELECT)
//		 { SimpleNode proj = (SimpleNode) node.jjtGetChild(1);
//		   	updateStatus(newStat, proj, env);
//		 }
	     /* original code by Patrick before May 15, 2002
		 if(node.type != OP_PROJECTOR && node.type != OP_SELECT)
		 { for(int i=0; i<node.jjtGetNumChildren(); i++)
		   { updateStatus(newStat, (SimpleNode) node.jjtGetChild(i), env);
		   }
		 }//else clause added by Yi Zheng March 30, 2002
		 else if(DEBUG)
		 { 
		 	if(node.type == OP_PROJECTOR)
		 		System.out.println("OP_PROJECTOR in updateStatus ignored: " + node.name);
		 	else if(node.type == OP_SELECT)
		 		System.out.println("OP_SELECT in updateStatus ignored: " + node.name);
		 	else
		 		throw new InterpretError("unexpected node type");
		}*/
	  }
   }   
} /*class CompBlock */
