
import java.util.*;

/**
 * Class CompTable is used to store computation objects that are
 * in a column of some relation. A surrogate is used to identify
 * the individual computations.
 */
public class CompTable
{     private Hashtable table;

public CompTable()
   {  table = new Hashtable();
   }   
   /**
	* Add an entry to the table. This is the version which is available
	* to the rest of the system for general use.
	*/
   public void addEntry(long surrogate, Computation comp)
   {  Long key = new Long(surrogate);
	  table.put(key, comp);
   }   
   /**
	* Add an entry to the table. This is called by the load method
	* of ExprTable when the system is started so as to initialize
	* the table from the .expr disk file.
	* @param   name   A string representation of the surrogate
	* @param   comp   The computation to put in the CompTable
	*/
   public void addEntry(String name, Computation comp)
   {  Long key = new Long(name);
	  table.put(key, comp);
   } 
     
   /**
	* Add an entry to the table. This is called (1) by the load method
	* of ExprTable when the system is started so as to initialize
	* the table from the .expr disk file.(2) by applyIjoin in computation
	* to export nested computations to callEnv
	* @param   fullName   A string representing the full path name of a computation
	* @param   surrogate   The surrogate value of the computation
	* @author Yi Zheng, May 14, 2002
	*/
	public void addEntry(String fullName, Long surrogate)
	{
		table.put(fullName, surrogate);
	}

   /**
	* Fill the exprtable (a hashtable) before dumping the expressions
	* to .expr disk file.
	* The key that is used is a string representation of the surrogate
	* prefixed by the symbol '#'. (This explains the name parameter of
	* the method addEntry() above)
	*
	* modified by Yi Zheng May 14, 2002
	*/
   public void fillExprTable(Hashtable exprTable)
   {  
		Hashtable surrTbl = new Hashtable();
		Hashtable nameTbl = new Hashtable();
		String name;
		try{
	   		//step 1, separate Long keys (surrogate) and String keys (fullName)
			for(Enumeration e = table.keys(); e.hasMoreElements();)
			{
				Object key = e.nextElement();
				if (key instanceof Long)
				{
					Computation c = (Computation) table.get((Long) key);
					surrTbl.put((Long) key, c);
				}
				else if (key instanceof String)
				{
					Long s = (Long) table.get((String) key);
					nameTbl.put((String) key, s);
				}
				else 
					throw new InterpretError("Invalid key type in exprTable");
			}
			
			//step 2, make sure there are equal number of surrogates and fullNames
			if (surrTbl.size() != nameTbl.size())
				throw new InterpretError("size mismatch between surrogates and computation names");
			
			//step 3, loop through surrTbl, add entries to exprTable
			for (Enumeration e = surrTbl.keys(); e.hasMoreElements();)
			{
				Long key = (Long) e.nextElement();
				Computation comp = (Computation) surrTbl.get(key);
				StringBuffer nb = new StringBuffer("#" + key.longValue() + "$");
				String fName = comp.parentName.concat("!" + comp.name);
				nb.append(fName);
				name = nb.toString();
				exprTable.put(name, new ExprEntry(name, (Object) comp.tree));
			}
			
		}catch (Exception e)
		{ Global.processThrowable(e);
		}

   }   

/* original code prior to Yi's modification above
   public void fillExprTable(Hashtable exprTable)
   {  String name;
	  for(Enumeration e = table.keys(); e.hasMoreElements();)
	  {  Long key = (Long) e.nextElement();
		 Computation comp = (Computation) table.get(key);
		 name = "#"+key.longValue(); 
//System.out.println("In CompTable, comp.tree.active = " + comp.tree.active);
		 exprTable.put(name, new ExprEntry(name, (Object) comp.tree));
	  } 
   }   
   */
   public Computation lookup(long surrogate)
   { return (Computation) table.get(new Long(surrogate));
   }   
   
   /**
    * look up the surrogate value of a computation by its full name
    * @param fullName fully qualified computation name, e.g. BANK!ba
    * @author Yi Zheng, May 14, 2002
    */
   public long lookup(String fullName)
   { 
     if (table.containsKey(fullName))
   		return ((Long) table.get(fullName)).longValue();
   	else return (long) -1;
   }
   public void rmComps(String name)
   {
   }   

   public Hashtable getCompTable() {
	return table;
   }
}
