

import java.util.*;
import java.io.*;

/**
 * Class for storing and accessing .dom system relation in memory.
 * Some information from the .rd system table is kept in class DomTable
 * as well. This is the information describing what attributes each
 * domain is defined on. This is for nested relations.
 * The .rd relation is described elsewhere.
 *
 * The .dom system relation is as follows:
 * .dom(.name  .type  .numref)<br>
 * where .name     identifier for the domain.<br>
 *       .type     the type of the domain: intg, strg, bool, or rel.<br>
 *       .numref   a reference count, i.e. how many time this domain is
 *                 referenced.<br>
 * @see RelTable
 */

public class DomTable implements TypeConstants, Constants
{
	//debugging Yi Zheng
	private boolean DEBUG=false;
   /**
	* Constants declaration.
	*/
   private static final int BUF_SIZE = 16 * 1024;
   private static final String DOM_NAME = ".dom";
   private static final String RD_NAME = ".rd";

   /**
	* A hashtable storing all the domains.
	*/
   private Hashtable table;

   private RelTable reltable;
   private Environment env;


   /**
	* Initialize the hash table, and load domain information
	* from disk files .dom & .rd.
	*/
   public DomTable(Environment env, boolean load)
   {  this.env = env;
	  reltable = env.reltable;
	  table = new Hashtable();
	  if(load){
		 this.load();
		 // showTable(true);
	  }
   }      
   /**
	* Add a domain object to this domtable.
	*/
   public void addDom(Domain d)
   {  table.put(d.name, d);
   }      

   /**
	* Dump this domain table. Information is seperated to .dom and .rd
	*/
/*
   public void dump(){
	  // dump to .dom and .rd files.
	  try{
	BlockOutputStream domos = new BlockOutputStream(DOM_NAME, BUF_SIZE);
	BlockOutputStream rdos = new BlockOutputStream(RD_NAME, BUF_SIZE);

	for(Enumeration e = table.keys(); e.hasMoreElements();){
	   Domain doment = (Domain) table.get(e.nextElement());
	   domos.writeString(doment.name);
	   domos.writeInt(doment.type);
	   domos.writeInt(doment.numref);

	   // for domain with 'idlist' type.. We don't need to take
	   // care of this since the dump() in RelTable will do its work.
	}
	// close the output file streams.
	domos.close();
	rdos.close();

	  }catch(IOException e){
	// Shouldn't happen..
	Global.println("error: domTable: dump IOException");

	// There must an error we didn't catch. If control comes here,
	// we must figure out what was going on and fix it.
	Global.exit(1);
	  }
   }         
   */


   /**
	* Dump this domain table. Information is seperated to .dom and .rd
	*	-- modified by Weizhong Sun, Jun 14th 1999
	*/
   public void dump(){
	  // dump to .dom  files.
	  try{
	BlockOutputStream domos = new BlockOutputStream(DOM_NAME, BUF_SIZE);
	BlockOutputStream rdos = new BlockOutputStream(RD_NAME, BUF_SIZE);//This is never used? Yi Zheng March 22, 2002
	String wb = new String("");

	for(Enumeration e = table.keys(); e.hasMoreElements();){
	   Domain doment = (Domain) table.get(e.nextElement());
	   domos.writeMyBytes(STRING, doment.name);
//	   domos.writeBytes("^F");
	   domos.writeByte(6);
	   wb = wb.valueOf(doment.type);
	   domos.writeMyBytes(STRING, wb);
//	   domos.writeBytes("^F");
	   domos.writeByte(6);
	   wb = wb.valueOf(doment.numref);
	   domos.writeMyBytes(STRING, wb);
//	   domos.writeBytes("^F");	// delimiter
	   domos.writeByte(6);
	   
	   //added by Yi Zheng April 29, 2002
	   wb = wb.valueOf(doment.isStateDom);
	   domos.writeMyBytes(STRING, wb);
	   domos.writeByte(6);

	   //added by Yi Zheng May 10, 2002
	  // wb = wb.valueOf(doment.surr);
	  // domos.writeMyBytes(STRING, wb);
	  // domos.writeByte(6);
	   //end of Yi's addition
	   
	   domos.writeBytes("\n");	// new line

	   // for domain with 'idlist' type.. We don't need to take
	   // care of this since the dump() in RelTable will do its work.
	}
	// close the output file streams.
	domos.close();
	rdos.close();

	  }catch(IOException e){
	// Shouldn't happen..
	Global.println("error: domTable: dump IOException");

	// There must an error we didn't catch. If control comes here,
	// we must figure out what was going on and fix it.
	Global.exit(1);
	  }
   }         



   /**
	* Fill the exprtable (a hashtable) before dumping the expressions
	* to .expr disk file.
	*/
   public void fillExprTable(Hashtable exprTable)
   {  String name;

	  for(Enumeration e = table.keys(); e.hasMoreElements();)
	  {  Domain doment = (Domain) table.get(e.nextElement());
		 if(doment.tree != null)
		 {  name = "."+doment.name; 
			exprTable.put(name, new ExprEntry(name, doment.tree));
		 }
	  } 
   }      
   /**
	* Get the reference to Domain object of the domain 'name'.
	*/
   public Domain getDom(String name){
	  return((Domain) table.get(name));
   }      
   /**
	* Get the attributes that a domain is defined on.
	*/
   public Vector getDoms(String domname){

	   // Check if this is a valid domain name.
	   Domain d = (Domain) table.get(domname);
	   if(d == null){
//	   	if(DEBUG)
//	   		System.out.println("DomTable:getDoms: " + 
//	   		"domain " + domname + " not in domTable");
	 return null;
	   }else{
	 // Check if this domain is of 'IDLIST' type.
	 if(d.type != IDLIST && d.type != COMP){
		return null;
		  }else {							
		  	
//		  	if(DEBUG)
//		  	{
//		  		System.out.println("DomTable:getDoms: " +
//		  		"domain " + domname + " is of type " + 
//		  		(d.type==IDLIST?"IDLIST":"COMP"));
//		  	}
		//	 Relation r = reltable.getRel("."+domname);//modified by Yi Zheng March 20, 2002
		Relation r = null;
		try
		{r = env.lookupRel("."+domname, true);}
		catch (InterpretError e)
		{ Global.processThrowable(e);}
		
		if(DEBUG)
		{
			if (r != null) 
				System.out.println("domain " + domname +
				"corresponds to relation " + r.name);
		}
		// Check if this nested domain is in reltable.
		if(r == null){
				// REVISIT: System error. There's no entry for this
	  // 'IDLIST'-typed domain. ..must be something wrong..
				//outstream.println("System error: There's no"+
	  //   " entry for the 'IDLIST'-typed domain "+domname);

	  // This is the case for a virtual domain with type
	  // 'IDLIST', e.g. domain R(a,b); let x be R;
	  return null;
			 }else if (r.domains != null){
//			 	if(DEBUG)
//			 		System.out.println("Copying over domlist");
		   Vector domlist = new Vector();
	  for(int i = 0; i < r.domains.length; i++){
		 domlist.addElement(r.domains[i].name);
				}
	  return domlist;
			}else return null;
		 } 
		 	
	  }
   }      
   /**
	* Get the type of a domain. 
	*/
//REVISIT: What is going on here???
   public int getType(String name)
	 throws InterpretError{
	  Domain d = (Domain) table.get(name);

	  // Only exception is in the case 'let x be S ijoin R;' where R is
	  // a top-level relation.
	  if(d == null){
	Relation r = reltable.getRel(name);
		 if(r != null){
			if(r.rvc == COMPUTATION)  return COMP;
			else return IDLIST;
		 }else
	   throw new InterpretError("domain "+name+" not found.");
	  }
	  return d.type;
   }      
   public Enumeration keys()
   {  return table.keys();
   }      


   /**
	* Load domain information from disk files(.dom).
	*/
/*
   private void load()
   {
   	String name;
	  int type;
	  int numref;

	  try
	  {  BlockInputStream domis = new BlockInputStream(DOM_NAME, BUF_SIZE); 
		 
		 // load data to hash table.
		 while(true)
		 {  // REVISIT: will need to handle bad record exceptions..
			try
			{  name = domis.readString();
			   type = domis.readInt();
			   numref = domis.readInt();
		  
			   addDom(new Domain(name, type, numref, null));
			}catch(IOException e)
			{  domis.close();   // EOF
			   break;
			}
		 }
	  }catch(FileNotFoundException e)
	  {  // REVISIT: not only here, but whenever the .dom is found to be
		 // corrupted.
		 // first time initialization.
		 addDom(new Domain(".rel_name", STRING, 2, null));
		 addDom(new Domain(".tuples", INTEGER, 1, null));
		 addDom(new Domain(".attributes", INTEGER, 1, null));
		 addDom(new Domain(".rvc", INTEGER, 1, null));
		 addDom(new Domain(".sort", INTEGER, 1, null));
		 addDom(new Domain(".dom_name", STRING, 2, null));
		 addDom(new Domain(".type", INTEGER, 1, null));
		 addDom(new Domain(".count", INTEGER, 1, null));
		 addDom(new Domain(".position", INTEGER, 1, null));
		 addDom(new Domain(".id", IDLIST, 1, null));
		 addDom(new Domain(".bool", BOOLEAN, 1, null));
		 addDom(new Domain(".short", SHORT, 1, null));
		 addDom(new Domain(".integer", INTEGER, 1, null));
		 addDom(new Domain(".long", LONG, 1, null));
		 addDom(new Domain(".float", FLOAT, 1, null));
		 addDom(new Domain(".double", DOUBLE, 1, null));
		 addDom(new Domain(".string", STRING, 1, null));
		 addDom(new Domain(".text", TEXT, 1, null));
		 addDom(new Domain(".expr", EXPR, 1, null));
		 addDom(new Domain(".stmt", STMT, 1, null));
	  }catch(IOException e)
	  {  Global.println("error: DomTable: load() IOException");
	  }
   }         
   */


   /**
	* Load domain information from disk files(.dom).
	*	-- modified by Weizhong Sun, Jun 14 1999
	*/
   private void load()
   {
   	String name;
	  int type;
	  int numref;
	  int isStateDom;	//Yi Zheng April 29, 2002
	//  long surr;		//Yi Zheng May 10, 2002
	  String rb;
	  Relation rel = new Relation();	// to call strRead()
	  Integer tint;
	 // Long tlong;		//Yi Zheng May 10, 2002

	  try
	  { 
	  	BlockInputStream domis = new BlockInputStream(DOM_NAME, BUF_SIZE); 
		 
		 // load data to hash table.
		 while(true)
		 {  // REVISIT: will need to handle bad record exceptions..
			try
			{ 
			   rb = rel.strRead(domis, STRING);
			   name = rb;
			   rb = rel.strRead(domis, INTEGER);
			   tint = new Integer(rb);
			   type = tint.intValue();
			   rb = rel.strRead(domis, INTEGER);
			   tint = new Integer(rb);
			   numref = tint.intValue();
			   
			   //added by Yi Zheng April 29, 2002
			   rb = rel.strRead(domis, INTEGER);
			   tint = new Integer(rb);
			   isStateDom = tint.intValue();
		  	
			   //added by Yi Zheng May 10, 2002
			  // rb = rel.strRead(domis, LONG);
			  // tlong = new Long(rb);
			  // surr = tlong.longValue();

		  		Domain nDomain = new Domain(name, type, numref, null);
		  		if (isStateDom == 1)
		  			nDomain.isStateDom = 1;
				else
					nDomain.isStateDom = 0;
		  		
			//	nDomain.surr = surr;
				//end Yi's addition	
		  		
			   addDom(nDomain);
			   
			}catch(IOException e)
			{  domis.close();   // EOF
			   break;
			}
		 }
	  }catch(FileNotFoundException e)
	  {  // REVISIT: not only here, but whenever the .dom is found to be
		 // corrupted.
		 // first time initialization.
		 addDom(new Domain(".rel_name", STRING, 2, null));
		 addDom(new Domain(".tuples", INTEGER, 1, null));
		 addDom(new Domain(".attributes", INTEGER, 1, null));
		 addDom(new Domain(".rvc", INTEGER, 1, null));
		 addDom(new Domain(".sort", INTEGER, 1, null));
		 addDom(new Domain(".dom_name", STRING, 2, null));
		 addDom(new Domain(".type", INTEGER, 1, null));
		 addDom(new Domain(".count", INTEGER, 1, null));
		 addDom(new Domain(".position", INTEGER, 1, null));
		 addDom(new Domain(".id", IDLIST, 1, null));
		 addDom(new Domain(".bool", BOOLEAN, 1, null));
		 addDom(new Domain(".short", SHORT, 1, null));
		 addDom(new Domain(".integer", INTEGER, 1, null));
		 addDom(new Domain(".long", LONG, 1, null));
		 addDom(new Domain(".float", FLOAT, 1, null));
		 addDom(new Domain(".double", DOUBLE, 1, null));
		 addDom(new Domain(".string", STRING, 1, null));
		 addDom(new Domain(".text", TEXT, 1, null));
		 addDom(new Domain(".expr", EXPR, 1, null));
		 addDom(new Domain(".stmt", STMT, 1, null));
		//Yi Zheng April 29, 2002
		 addDom(new Domain(".isState", INTEGER, 1, null));

	  }catch(IOException e)
	  {  Global.println("error: DomTable: load() IOException");
	  }
   }         



   /**
	* Delete a domain from the DomTable.
	*/
   public boolean rmDom(String name)
   {  // look for the domain we want to remove.
	  Domain d = (Domain) table.get(name);

	  // if domain exists and is not being referenced...
	  if(d != null && d.numref == 0)
	  {  // decrement the reference counter of referenced domains
		 // when necessary.
		 if(d.type == IDLIST || d.type == COMP)
		 {  // remove this entry in relTable, which will decrement the
			// reference counter of the domains in this relation.
			reltable.rmRel("."+name); 
		 }
		 // remove this domain from hashtable.
		 table.remove(name);
		 return true;
	  }
	  else if(d == null)
	  {  // this is not a declared domain.
		 Global.println("warning : domain '" + name + "' is not found in domTable.");
		 return true;
	  }else
	  {  // this domain is being referenced.
		 Global.println("warning: domain '"+name+"' is being referenced.");
		 return false;
	  }
   }         
   /**
	* Print the information for one domain.
	* @param doment the domain whose information we want to print.
	*/
   public void showDom(Domain doment)
   {  String domlistString = ""; 
	  PrettyPrint out = new PrettyPrint();

	  // print the title.
	  out.pprintRule(31);    out.pprint(" Domain Entry ", 14);
	  out.pprintRule(31);    out.println();
//	  out.pprint("Name        Type            NumRef       Dom_List", 55); Yi Zheng April 29, 2002
	  out.pprint("Name        Type            NumRef     IsState   Dom_List", 65);
	  out.println();
	  out.pprintRule(76);
	  out.println();

	  // print name, type and numref etc.
	  out.pprint(doment.name,12);
	  out.pprint(Utility.type2String(doment.type), 16);
	  out.pprint(doment.numref,11);

		//Yi Zheng April 29, 2002, show whether the domain is a hidden state domain
		//if a domain is a state domain, it cannot be used in projection list etc. i.e. it's
		//invisible to the top level user
		byte b;
		if (doment.isStateDom == 1)
			b = (byte) 1;
		else
			b = (byte) 0;
		out.pprint(b, 10);
		//end Yi's addition
		
		
	  // print domain list if there is any, i.e. IDLIST type.
	  Vector domlist = getDoms(doment.name);
	  if(domlist != null)
	  {  for(int j = 0; j < domlist.size(); j++)
		 {   domlistString += (String)domlist.elementAt(j)+ ", ";
		 }
		 out.pprint(domlistString, 24);
		 domlistString = "";        // reset the buffer.
	  }

	  // print syntax tree if virtual domain.
	  if(doment.tree != null)
	  {  out.println(); 
		 if(doment.tree.jjtGetNumChildren() == 1)
			// virtual domain without 'initial' tree.
			((SimpleNode)(doment.tree.jjtGetChild(0))).dump(
			"										");
		 else  
			// virtual domain with 'initial' tree.
			((SimpleNode)(doment.tree.jjtGetChild(1))).dump(
			"                              			");
	  }
	  out.println();
	  out.pprintRule(76);
	  out.println();
   }         
   /**
	* Print the domain table.
	* @param flag true if only print user-defined domains info.;
	*             false if want print both system domain and user-defined 
	*             domains.
	*/
   public void showTable(boolean flag)
   {  String domlistString = "";
	  Domain doment;
	  PrettyPrint out = new PrettyPrint();

	  // print the title.
	  out.pprintRule(31);    out.pprint(" Domain Entry ", 14);
	  out.pprintRule(31);    out.println();
//	  out.pprint("Name        Type            NumRef       Dom_List", 55); Yi Zheng April 29, 2002
	  out.pprint("Name        Type            NumRef     IsState   Dom_List", 65);
	  out.println();
	  out.pprintRule(76);
	  out.println();

	  for(Enumeration e = table.keys(); e.hasMoreElements(); )
	  {  doment = (Domain) table.get(e.nextElement());

		 if(flag || !doment.name.startsWith("."))
		 {  // print name, type and numref etc.
			out.pprint(doment.name,12);
			out.pprint(Utility.type2String(doment.type),16);
			out.pprint(doment.numref,11);
			//Yi Zheng April 29, 2002, show whether the domain is a hidden state domain
			//if a domain is a state domain, it cannot be used in projection list etc. i.e. it's
			//invisible to the top level user
			byte b;
			if (doment.isStateDom == 1)
				b = (byte) 1;
			else
				b = (byte) 0;
			out.pprint(b, 10);
			//end Yi's addition

			// print domain list if there is any, i.e. IDLIST type.
			Vector domlist = getDoms(doment.name);
			if(domlist != null)
			{  for(int j = 0; j < domlist.size(); j++)
			   {  domlistString += (String)domlist.elementAt(j)+ ", ";
			   }
			   out.pprint(domlistString, 24);
			   domlistString = "";        // reset the buffer.
			}

			// print syntax tree if virtual domain.
			SimpleNode tree = env.lookupDom(doment.name, false).getTree();
			if(tree != null)
			{  out.println();
			   tree.dump("										");
			   out.pprint("--");
			}
			out.println();
		 }
	  }
	  out.pprintRule(76);
	  out.println();
   }  
   
   /**
    * function to provide the size of the domTable
    */
    public int getDomTableSize()
    { return table.size();}       
}
